def drilling_influx_1(tvd, rho_fluid, rho_mud, frac_gradient):
    """
    Calculate differential pressure profile with influx during drilling case 1 (fracture pressure at
    shoe and gas gradient above - one fluid behind the casing)
    :param tvd: list - true vertical depth, m
    :param rho_fluid: rho_fluid: fluid density, sg
    :param rho_mud: mud density, sg
    :param frac_gradient: fracture gradient, bar/m
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import frac_shoe_gas_grad_above
    from .pressure_external import onefluid_behindcasing

    p_int = frac_shoe_gas_grad_above(tvd, frac_gradient, rho_fluid)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def drilling_influx_2(tvd, rho_mud, tvd_next_section, fraction=0.5):
    """
    Calculate differential pressure profile with influx during drilling case 2 (fraction of bhp at wh -
    one fluid behind the casing)
    :param tvd: list - true vertical depth, m
    :param rho_mud: mud density, sg
    :param tvd_next_section: tvd at bottom, m
    :param fraction: bhp fraction at wh
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import fraction_of_bhp_at_wh
    from .pressure_external import onefluid_behindcasing

    p_int = fraction_of_bhp_at_wh(tvd, rho_mud, tvd_next_section, fraction)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def drilling_influx_3(tvd, rho_mud, id_csg, od_dp, p_res, tvd_res, vol_kick_initial):
    """
    Calculate differential pressure profile with influx during drilling case 3 (gas kick profile -
    one fluid behind the casing)
    :param tvd: list - true vertical depth, m
    :param rho_mud: mud density, sg
    :param id_csg: casing inner diameter, in
    :param od_dp: drill pipe outer diameter, in
    :param p_res: reservoir pressure, bar
    :param tvd_res: tvd at reservoir, m
    :param vol_kick_initial: influx initial volume, m3
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import gas_kick
    from .pressure_external import onefluid_behindcasing

    p_int = gas_kick(tvd, rho_mud, p_res, tvd_res, vol_kick_initial, id_csg, od_dp)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def pressure_test_onefluid(tvd, p_test, rho_fluid_int, rho_fluid_ext):
    """
    Calculate differential pressure profile during pressure testing with one fluid behind the casing.
    :param tvd: list - true vertical depth, m
    :param p_test: testing pressure, bar
    :param rho_fluid_int: float - internal fluid density, sg
    :param rho_fluid_ext: float - external fluid density, sg
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import pressure_test
    from .pressure_external import onefluid_behindcasing

    p_int = pressure_test(tvd, p_test, rho_fluid_int)
    p_ext = onefluid_behindcasing(tvd, rho_fluid_ext)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def pressure_test_morefluids(tvd, p_test, rho_mud, rho_fluid, tvd_fluid):
    """
    Calculate differential pressure profile during pressure testing with more than one fluid behind the casing.
    :param tvd: list - true vertical depth, m
    :param p_test: testing pressure, bar
    :param rho_mud: list - downwards sorted mud densities, sg
    :param rho_fluid: list - downwards sorted fluids densities, sg
    :param tvd_fluid: list - reference tvd of fluid change, m
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import pressure_test
    from .pressure_external import morefluids_behindcasing

    p_int = pressure_test(tvd, p_test, rho_mud)
    p_ext = morefluids_behindcasing(tvd, rho_fluid, tvd_fluid)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def production_with_packer(tvd, rho_fluid, rho_mud, p_res, tvd_perf, rho_packerfluid, tvd_packer):
    """
    Calculate differential pressure profile during production with packer.
    :param tvd: list - true vertical depth, m
    :param rho_fluid: fluid density, sg
    :param rho_mud: mud density, sg
    :param p_res: reservoir pressure, bar
    :param tvd_perf: tvd at perforations, m
    :param rho_packerfluid: packer fluid density, sg
    :param tvd_packer: tvd at packer, m
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import tubing_leak
    from .pressure_external import onefluid_behindcasing

    p_int = tubing_leak(tvd, p_res, rho_fluid, tvd_perf, rho_packerfluid, tvd_packer, rho_mud)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def production_with_packer_and_depletedzone(tvd, rho_fluid, rho_mud, p_res, tvd_perf, rho_packerfluid, tvd_packer,
                                            tvd_zone, p_zone):
    """
    Calculate differential pressure profile during production with packer and a depleted zone.
    :param tvd: list - true vertical depth, m
    :param rho_fluid: fluid density, sg
    :param rho_mud: mud density, sg
    :param p_res: reservoir pressure, bar
    :param tvd_perf: tvd at perforations, m
    :param rho_packerfluid: packer fluid density, sg
    :param tvd_packer: tvd at packer, m
    :param tvd_zone: tvd at depleted zone, m
    :param p_zone: pressure at depleted zone, bar
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import tubing_leak
    from .pressure_external import depleted_zone

    p_int = tubing_leak(tvd, p_res, rho_fluid, tvd_perf, rho_packerfluid, tvd_packer, rho_mud)
    p_ext = depleted_zone(tvd, tvd_zone, p_zone, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def gas_filled(tvd, p_res, rho_gas, tvd_res, rho_mud):
    """
    Calculate differential pressure profile when the entire casing string is filled with gas.
    :param tvd: list - true vertical depth, m
    :param p_res: reservoir pressure, bar
    :param rho_gas: gas density, sg
    :param tvd_res: tvd at reservoir, m
    :param rho_mud: mud density, sg
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import displacement_to_gas
    from .pressure_external import onefluid_behindcasing

    p_int = displacement_to_gas(tvd, p_res, rho_gas, tvd_res)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def production_without_packer_and_depletedzone(tvd, p_res, rho_gas, tvd_res, tvd_zone, p_zone, rho_mud):
    """
    Calculate differential pressure profile during production with depleted zone and without packer.
    :param tvd: list - true vertical depth, m
    :param p_res: reservoir pressure, bar
    :param rho_gas: gas density, sg
    :param tvd_res: tvd at reservoir, m
    :param tvd_zone: tvd at depleted zone, m
    :param p_zone: pressure at depleted zone, bar
    :param rho_mud: mud density, sg
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import displacement_to_gas
    from .pressure_external import depleted_zone

    p_int = displacement_to_gas(tvd, p_res, rho_gas, tvd_res)
    p_ext = depleted_zone(tvd, tvd_zone, p_zone, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def stimulation(tvd, whp, rho_injectionfluid, rho_mud, rho_packerfluid, tvd_packer=0):
    """
    Calculate differential pressure profile during stimulation.
    :param tvd: list - true vertical depth, m
    :param whp: wellhead pressure, bar
    :param rho_injectionfluid: injection fluid density, sg
    :param rho_mud: mud density, sg
    :param rho_packerfluid: packer fluid density, sg
    :param tvd_packer: tvd at packer, m
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import tubing_leak_stimulation
    from .pressure_external import onefluid_behindcasing

    p_int = tubing_leak_stimulation(tvd, whp, rho_packerfluid, rho_injectionfluid, tvd_packer)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def fluid_storage_onefluid_behindcasing(tvd, rho_fluid, frac_gradient, rho_mud):
    """
    Calculate differential pressure profile during fluid storage with one fluid behind the casing.
    :param tvd: list - true vertical depth, m
    :param rho_fluid: fluid density, sg
    :param frac_gradient: fracture gradient, bar/m
    :param rho_mud: mud density, sg
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import frac_shoe_gas_grad_above
    from .pressure_external import onefluid_behindcasing

    p_int = frac_shoe_gas_grad_above(tvd, frac_gradient, rho_fluid)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def fluid_storage_morefluids_behindcasing(tvd, frac_gradient, rho_fluid, tvd_fluid):
    """
    Calculate differential pressure profile during fluid storage with more than one fluid behind the casing.
    :param tvd: list - true vertical depth, m
    :param rho_fluid: fluid density, sg
    :param frac_gradient: fracture gradient, bar/m
    :param rho_fluid: list - downwards sorted fluids densities, sg
    :param tvd_fluid: list - reference tvd of fluid change, m
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import frac_shoe_gas_grad_above
    from .pressure_external import morefluids_behindcasing

    p_int = frac_shoe_gas_grad_above(tvd, frac_gradient, rho_fluid)
    p_ext = morefluids_behindcasing(tvd, rho_fluid, tvd_fluid)

    pressure_differential = p_int - p_ext

    return pressure_differential


def fluid_storage_depletedzone(tvd, frac_gradient, rho_fluid, tvd_zone, p_zone, rho_mud):
    """
    Calculate differential pressure profile during fluid storage with depleted zone.
    :param tvd: list - true vertical depth, m
    :param frac_gradient: fracture gradient, bar/m
    :param rho_fluid: fluid density, sg
    :param tvd_zone: tvd at depleted zone, m
    :param p_zone: pressure at depleted zone, bar
    :param rho_mud: mud density, sg
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import frac_shoe_gas_grad_above
    from .pressure_external import depleted_zone

    p_int = frac_shoe_gas_grad_above(tvd, frac_gradient, rho_fluid)
    p_ext = depleted_zone(tvd, tvd_zone, p_zone, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential



from numpy import interp


def gen_msgs(pipe):
    necessary_inputs = {'Displacement to gas': {'resPressure': 'Reservoir pressure is missing',
                                                'resTvd': 'Reservoir depth (tvd) is missing'},
                        'Production': {'resPressure': 'Reservoir pressure is missing',
                                       'packerTvd': 'Packer depth (tvd) is missing',
                                       'perforationsTvd': 'Depth (tvd) of perforations is missing'},
                        'Injection': {'whp': 'Wellhead Pressure during injection is missing',
                                      'injectionFluid': 'Injection fluid density is missing'},
                        'Pressure Test': {'testFluidDensity': 'Fluid density for pressure test is missing',
                                          'testPressure': 'Testing pressure is missing',
                                          'pipeDiameter': 'Effective pipe diameter is missing'},
                        'Gas Kick': {'resPressure': 'Reservoir pressure is missing',
                                     'resTvd': 'Reservoir depth (tvd) is missing'}}

    missing_loads = {}

    for load, reference in necessary_inputs.items():
        msgs = []
        for parameter, msg in reference.items():
            status = 0

            for section in pipe.settings.values():
                if parameter in section:
                    status = 1

            if status == 0:         # parameter was not found in settings
                msgs.append(msg)

        if len(msgs) > 0:
            missing_loads[load] = msgs

    pipe.msgs = missing_loads


def define_max_loads(loads):
    for load in loads:
        min_level = {'force': min(load['axialForce']), 'pressure': min(load['diffPressure'])}
        max_level = {'force': max(load['axialForce']), 'pressure': max(load['diffPressure'])}
        max_loads = {}

        if min_level['force'] < 0:
            max_loads['compression'] = min_level['force']
        else:
            max_loads['compression'] = None

        if max_level['force'] > 0:
            max_loads['tension'] = max_level['force']
        else:
            max_loads['tension'] = None

        if min_level['pressure'] < 0:
            max_loads['collapse'] = min_level['pressure']
            load['_MaxCollapsePoint'] = load['axialForce'][load['diffPressure'].index(min_level['pressure'])]
        else:
            max_loads['collapse'] = None

        if max_level['pressure'] > 0:
            max_loads['burst'] = max_level['pressure']
        else:
            max_loads['burst'] = None

        load['maxLoads'] = max_loads


def get_collapse_base(csg, axial_force):
    if axial_force <= 0:
        collapse_base = csg.limits['collapse']
    else:
        collapse_base = interp(axial_force, csg.ellipse[0], csg.ellipse[2])
    return collapse_base


def define_min_df(csg):
    base = {'compression': csg.conn_limits[0], 'tension': csg.conn_limits[1], 'burst': csg.limits['burst']}

    for load in csg.loads:
        if '_MaxCollapsePoint' in load:
            base['collapse'] = get_collapse_base(csg, load['_MaxCollapsePoint'])
        load['minDF'] = {}
        for load_type, value in load['maxLoads'].items():
            if value is None:
                load['minDF'][load_type] = None
            else:
                load['minDF'][load_type] = base[load_type] / value


def define_safety_factors(csg):
    precaution = {'burst': None, 'collapse': None, 'tension': None, 'compression': None}
    for load_type in precaution.keys():
        warning = {'load': None, 'safetyFactor': 1000, 'maxLoad': None}
        for load in csg.loads:
            if load['minDF'][load_type] is None:
                value = 1000
            else:
                value = load['minDF'][load_type]
            if value < warning['safetyFactor']:
                warning['load'] = load['description']
                warning['safetyFactor'] = round(load['minDF'][load_type], 2)
                warning['maxLoad'] = round(abs(load['maxLoads'][load_type]), 2)
                precaution[load_type] = warning

    csg.safety_factors = precaution



from .unit_converter import convert_unit


def gen_running(csg, tvd_fluid=None, rho_fluid=None, v_avg=0.3, fric=0.24, a=1.5):
    """
    Run load case: Running in hole

    Arguments:
        csg: casing obj
        tvd_fluid (list or None): reference tvd of fluid change
        rho_fluid (list or None): downwards sorted fluids densities
        v_avg (num): average running speed, m/s
        fric (num): sliding friction factor pipe - wellbore
        a (num): ratio of maximum running speed to average running speed

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import running

    if tvd_fluid is None:
        tvd_fluid = []
    if rho_fluid is None:
        rho_fluid = [1.2]

    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = running(csg.trajectory, csg.nominal_weight, csg.od, csg.id,
                                                 csg.shoe, tvd_fluid, rho_fluid, v_avg, e, fric, a)

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Running', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_overpull(csg, tvd_fluid=None, rho_fluid=None, v_avg=0.3, fric=0.24, a=1.5, f_ov=0.0):
    """
    Run load case: Overpull

    Arguments:
        csg: casing obj
        tvd_fluid (list or None): reference tvd of fluid change
        rho_fluid (list or None): downwards sorted fluids densities
        v_avg (num): average running speed, m/s
        fric (num): sliding friction factor pipe - wellbore
        a (num): ratio of maximum running speed to average running speed
        f_ov (int or num): overpull force (often during freeing of stuck pipe), kN.

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import overpull

    if tvd_fluid is None:
        tvd_fluid = []
    if rho_fluid is None:
        rho_fluid = [1.2]

    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = overpull(csg.trajectory, csg.nominal_weight, csg.od, csg.id,
                                                  csg.shoe, tvd_fluid, rho_fluid, v_avg, e, fric, a, f_ov)

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Overpull', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_green_cement(csg, rho_fluid_int=1.2, rho_cement=1.8, f_pre=0.0, p_test=0.0):
    """
    Run load case: Green Cement Pressure test

    Arguments:
        csg: casing obj
        rho_cement (num): cement density, sg
        rho_fluid_int (num): inside fluid density, sg
        f_pre (num): pre-loading force applied to the casing string if necessary, kN
        p_test (num): testing pressure, psi

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import green_cement_pressure_test

    f_test = convert_unit(p_test * csg.area, unit_from="lbf", unit_to="kN")
    p_test = convert_unit(p_test, unit_from="psi", unit_to="bar")
    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = green_cement_pressure_test(csg.trajectory, csg.nominal_weight,
                                                                    csg.od, csg.id, rho_cement, rho_fluid_int,
                                                                    p_test, e, f_test, f_pre)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Green Cement Pressure Test', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_cementing(csg, rho_cement=1.8, rho_fluid=1.3, f_pre=0.0):
    """
    Run load case: Cementing

    Arguments:
        csg: casing obj
        rho_cement (num): cement density, sg
        rho_fluid (num): displacement fluid density, sg
        f_pre (int or num): pre-loading force applied to the casing string if necessary, kN

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import cementing

    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = cementing(csg.trajectory, csg.nominal_weight, csg.od, csg.id,
                                                   rho_cement, rho_fluid, e, f_pre)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Cementing', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_displacement_gas(csg, p_res, tvd_res, rho_gas=0.5, rho_mud=1.4):
    """
    Run load case: Displacement to gas

    Arguments:
        csg: casing obj
        p_res (num): reservoir pressure, psi
        tvd_res (num): tvd at reservoir, m
        rho_gas (num): gas density, sg
        rho_mud (num): mud density, sg

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import gas_filled

    p_res = convert_unit(p_res, unit_from='psi', unit_to='bar')
    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = gas_filled(csg.trajectory, csg.nominal_weight, csg.od, csg.id,
                                                    rho_mud, rho_gas, p_res, tvd_res, e)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Displacement to gas', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_production(csg, p_res, rho_prod_fluid, rho_ann_fluid, rho_packerfluid, md_toc, tvd_packer, tvd_perf,
                   poisson=0.3, f_setting=0.0):
    from .load_cases import production_with_packer

    p_res = convert_unit(p_res, unit_from='psi', unit_to='bar')
    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = production_with_packer(csg.trajectory, md_toc, csg.od, csg.id,
                                                                rho_prod_fluid, rho_ann_fluid, e, p_res, tvd_perf,
                                                                rho_packerfluid, tvd_packer, poisson, f_setting)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Production', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_injection(csg, whp, rho_injectionfluid, rho_mud, temp, t_k, alpha=17e-6, poisson=0.3, f_setting=0):
    from .load_cases import stimulation

    whp = convert_unit(whp, unit_from='psi', unit_to='bar')
    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = stimulation(csg.trajectory, csg.toc_md, csg.od, csg.id, e,
                                                     whp, rho_injectionfluid, rho_mud, rho_mud, temp, t_k,
                                                     alpha=alpha, poisson=poisson, f_setting=f_setting)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Injection', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_full_evacuation(csg, rho_prod_fluid, rho_mud, md_toc, poisson=0.3, f_setting=0.0):
    from .load_cases import production_evacuation

    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = production_evacuation(csg.trajectory, csg.od, csg.id, md_toc, rho_prod_fluid,
                                                               rho_mud, e, poisson, f_setting)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Full Evacuation', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_pressure_test(csg, whp, effective_diameter, rho_testing_fluid, rho_mud):
    from .load_cases import pressure_test

    whp = convert_unit(whp, unit_from='psi', unit_to='bar')
    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = pressure_test(csg.trajectory, whp, csg.od, e, effective_diameter,
                                                       rho_testing_fluid, rho_mud)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Pressure Test', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_gas_kick(csg, p_res, tvd_res, rho_gas=0.5, rho_mud=1.4, vol_kick_initial=0.05):
    """
    Run load case: Gas Kick

    Arguments:
        csg: casing obj
        p_res (num): reservoir pressure, psi
        tvd_res (num): tvd at reservoir, m
        rho_gas (num): gas density, sg
        rho_mud (num): mud density, sg
        vol_kick_initial (num): influx initial volume, m3

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import gas_kick

    p_res = convert_unit(p_res, unit_from='psi', unit_to='bar')
    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = gas_kick(csg.trajectory, csg.nominal_weight, csg.od, csg.id, rho_mud, rho_gas,
                                                  p_res, tvd_res, e, vol_kick_initial)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Gas kick', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})


def gen_mud_drop(csg, rho_mud=1.4, rho_mud_new=1.1):
    """
    Run load case: Mud Drop

    Arguments:
        csg: casing obj
        rho_mud (num): mud density, sg
        rho_mud_new (num): new mud density, sg

    Returns:
        None. It adds the load case results in loads as [load case name, axial_force, pressure_differential]
    """

    from .load_cases import mud_drop

    e = convert_unit(csg.e, unit_from='psi', unit_to='bar')

    axial_force, pressure_differential = mud_drop(csg.trajectory, csg.nominal_weight, csg.od, csg.id, rho_mud,
                                                  rho_mud_new, e)

    pressure_differential = convert_unit(pressure_differential, unit_from="Pa", unit_to="psi")

    axial_force = [x * 1000 / 4.448 for x in axial_force]  # kN to lbf

    csg.loads.append({'description': 'Mud Drop', 'axialForce': axial_force,
                      'diffPressure': pressure_differential})



from math import pi
from .unit_converter import convert_unit
from .collapse_calcs import calc_collapse_pressure
from .von_mises import vme
from .design_factors import api_limits
from .connections import get_conn_limits
from .utilities import *
from .prepare_cases import *
import well_profile as wp


class Casing(object):
    """
    Casing object.

    Arguments:
        pipe (dict): set the main pipe characteristics. 'od', 'id', 'shoeDepth', 'tocMd',  'top', 'casingClass',
                     'weight'(opt), 'yield'(opt), 'e'(opt)
        factors (dict): set define factors for pipe and connection.

    Attributes:
        od (num): outer diameter of the casing [in]
        id (num): inner diameter of the casing [in]
        thickness (str or None): outer diameter of the casing [in]
        dt (num): ratio --> outer diameter / thickness
        area (num): effective area [in^2]
        shoe (num): measured depth at shoe [m]
        ellipse (list): triaxial points [x, y+, y-]
        loads (list): list of loads that have been run
        nominal_weight (num): weight per unit length [kg/m]
        trajectory (obj): wellbore trajectory object
        api_lines (list): API limits coordinates [x, y]
        design_factor (dict): design factors used 'vme', 'api'
    """

    def __init__(self, pipe, conn_compression=0.6, conn_tension=0.6, factors=None):

        df = {'pipe': {'tension': 1.1, 'compression': 1.1, 'burst': 1.1, 'collapse': 1.1, 'triaxial': 1.25},
              'connection': {'tension': 1.0, 'compression': 1.0}}

        if type(factors) == dict:
            for key in factors.keys():
                for item in factors[key].keys():
                    df[key][item] = factors[key][item]

        self.od = pipe['od']
        self.id = pipe['id']
        self.area = (pi / 4) * (self.od ** 2 - self.id ** 2)
        self.thickness = (self.od - self.id) / 2
        self.dt = self.od / self.thickness
        self.toc_md = pipe['tocMd']
        self.shoe = pipe['shoeDepth']
        self.top = pipe['top']

        if 'casingClass' in pipe:
            self.pipe_class = pipe['casingClass']
        else:
            self.pipe_class = None

        if 'yield' in pipe:
            yield_s = pipe['yield']
        else:
            yield_s = 80000

        if 'weight' in pipe:
            self.nominal_weight = pipe['weight']
        else:
            self.nominal_weight = 64

        if 'e' in pipe:
            self.e = pipe['e']
        else:
            self.e = 29e6

        self.limits = {'burst': 0.875 * 2 * yield_s * self.thickness / self.od,
                       'burstDF': 0.875 * 2 * yield_s * self.thickness / self.od / df['pipe']['burst'],
                       'collapse': - calc_collapse_pressure(self.dt, yield_s),
                       'collapseDF': - calc_collapse_pressure(self.dt, yield_s) / df['pipe']['collapse'],
                       'compression': - yield_s * self.area,
                       'compressionDF': - yield_s * self.area / df['pipe']['compression'],
                       'tension': yield_s * self.area,
                       'tensionDF': yield_s * self.area / df['pipe']['tension']}

        self.ellipse = vme(yield_s, self.area, self.id, self.od, df['pipe']['triaxial'])
        self.loads = []
        self.trajectory = None
        self.settings = None
        self.msgs = None
        self.safety_factors = None
        self.api_lines, self.collapse_curve = api_limits(self.dt, yield_s, self.limits, self.area,
                                                         df['pipe']['tension'],
                                                         df['pipe']['compression'],
                                                         df['pipe']['burst'],
                                                         df['pipe']['collapse'])
        self.conn_limits = get_conn_limits(self.limits, conn_compression, conn_tension,
                                           df['connection']['compression'],
                                           df['connection']['tension'])
        self.design_factor = {'vme': df['pipe']['triaxial'],
                              'api': {'compression': df['pipe']['compression'],
                                      'tension': df['pipe']['tension'],
                                      'burst': df['pipe']['burst'],
                                      'collapse': df['pipe']['collapse']}}

    def add_trajectory(self, survey):

        trajectory = wp.load(survey, equidistant=False)
        idx = [trajectory.trajectory.index(x) for x in trajectory.trajectory if self.top <= x['md'] <= self.shoe]
        trajectory.md = [x['md'] - self.top for x in trajectory.trajectory][idx[0]:idx[-1]+1]
        trajectory.tvd = [x['tvd'] - self.top for x in trajectory.trajectory][idx[0]:idx[-1]+1]
        trajectory.inclination = [x['inc'] for x in trajectory.trajectory][idx[0]:idx[-1]+1]
        trajectory.azimuth = [x['azi'] for x in trajectory.trajectory][idx[0]:idx[-1]+1]
        trajectory.dls = [x['dls'] for x in trajectory.trajectory][idx[0]:idx[-1]+1]
        self.trajectory = trajectory

    def plot(self, plot_type='plotly'):
        from .plot import create_plotly_figure, create_pyplot_figure
        if plot_type == 'plotly':
            fig = create_plotly_figure(self)
        else:
            fig = create_pyplot_figure(self)

        return fig
    
    def run_loads(self, settings=None):

        self.define_settings(settings)
        config = self.settings
        gen_msgs(self)

        gen_overpull(self, rho_fluid=[config['densities']['mud']], v_avg=config['tripping']['speed'],
                     fric=config['tripping']['slidingFriction'], a=config['tripping']['maxSpeedRatio'],
                     f_ov=int(config['forces']['overpull']))

        gen_running(self, rho_fluid=[config['densities']['mud']], v_avg=config['tripping']['speed'],
                    fric=config['tripping']['slidingFriction'], a=config['tripping']['maxSpeedRatio'])

        gen_green_cement(self, rho_fluid_int=config['densities']['cementDisplacingFluid'],
                         rho_cement=config['densities']['cement'], f_pre=config['forces']['preloading'],
                         p_test=config['testing']['cementingPressure'])

        gen_cementing(self, rho_cement=config['densities']['cement'],
                      rho_fluid=config['densities']['cementDisplacingFluid'],
                      f_pre=config['forces']['preloading'])

        if self.pipe_class in [None, 'Production']:
            gen_full_evacuation(self, rho_prod_fluid=config['production']['fluidDensity'],
                                rho_mud=config['densities']['mud'], md_toc=self.toc_md,
                                poisson=config['production']['poisson'],
                                f_setting=config['forces']['preloading'])

        gen_mud_drop(self, rho_mud=config['densities']['mud'], rho_mud_new=config['densities']['mudDropTo'])

        if 'Displacement to gas' not in self.msgs:
            gen_displacement_gas(self, p_res=config['production']['resPressure'],
                                 tvd_res=config['production']['resTvd'],
                                 rho_gas=config['densities']['gasKick'], rho_mud=config['densities']['mud'])

        if 'Production' not in self.msgs and self.pipe_class in [None, 'Production']:
            gen_production(self, p_res=config['production']['resPressure'],
                           rho_prod_fluid=config['production']['fluidDensity'],
                           rho_ann_fluid=config['densities']['completionFluid'],
                           rho_packerfluid=config['production']['packerFluidDensity'],
                           md_toc=self.toc_md,
                           tvd_packer=config['production']['packerTvd'],
                           tvd_perf=config['production']['perforationsTvd'],
                           poisson=config['production']['poisson'],
                           f_setting=config['forces']['preloading'])

        if 'Injection' not in self.msgs and self.pipe_class in [None, 'Production']:
            gen_injection(self, whp=config['injection']['whp'],
                          rho_injectionfluid=config['densities']['injectionFluid'],
                          rho_mud=config['densities']['mud'], temp=config['temp'],
                          t_k=config['production']['wellHeadTemp'],
                          poisson=config['production']['poisson'],
                          f_setting=config['forces']['preloading'])

        if 'Pressure Test' not in self.msgs:
            gen_pressure_test(self, whp=config['testing']['testPressure'],
                              effective_diameter=config['testing']['pipeDiameter'],
                              rho_testing_fluid=config['testing']['testFluidDensity'],
                              rho_mud=config['densities']['mud'])

        if 'Gas Kick' not in self.msgs:
            gen_gas_kick(self, p_res=config['production']['resPressure'], tvd_res=config['production']['resTvd'],
                         rho_gas=config['densities']['gasKick'], rho_mud=config['densities']['mud'],
                         vol_kick_initial=config['influx']['gasKickVolume'])

        define_max_loads(self.loads)
        define_min_df(self)
        define_safety_factors(self)

    def define_settings(self, settings):

        default = {'densities': {'mud': 1.5, 'cement': 1.9, 'cementDisplacingFluid': 1.6, 'gasKick': 0.5,
                                 'completionFluid': 1.8, 'injectionFluid': 1.3, 'mudDropTo': 1.1},
                   'tripping': {'slidingFriction': 0.24, 'speed': 0.3, 'maxSpeedRatio': 1.5},
                   'production': {'fluidDensity': 0.9, 'packerFluidDensity': 1.3, 'poisson': 0.3, 'wellHeadTemp': 5},
                   'forces': {'overpull': 0,
                              'preloading': 0},
                   'testing': {'cementingPressure': 1500},
                   'injection': {'whp': 2000, 'injectionFluid': 1.3},
                   'temp': {'seabed': {'tvd': 500, 'temp': 4},
                            'target': {'tvd': 1500, 'temp': 160}},
                   'influx': {'gasKickVolume': 20}}

        if type(settings) == dict:
            for key in settings.keys():
                for item in settings[key].keys():
                    default[key][item] = settings[key][item]

        self.settings = default



def calc_collapse_pressure(dt, yield_strength, axial_stress=None):

    y_p = yield_strength

    if axial_stress is not None:
        s_a = axial_stress
        y_pa = ((1 - 0.75 * (s_a / y_p) ** 2) ** 0.5 - 0.5 * s_a / y_p) * y_p
        y_p = y_pa

    if y_p != 0:
        a = 2.8762 + 0.10679 * 1e-5 * y_p + 0.21301 * 1e-10 * y_p ** 2 - 0.53132 * 1e-16 * y_p ** 3
        b = 0.026233 + 0.50609 * 1e-6 * y_p
        c = -465.93 + 0.030867 * y_p - 0.10483 * 1e-7 * y_p ** 2 + 0.36989 * 1e-13 * y_p ** 3
        f = (46.95 * 1e6 * ((3 * b / a) / (2 + b / a)) ** 3) / \
            (y_p * ((3 * b / a) / (2 + b / a) - (b / a)) * (1 - ((3 * b / a) / (2 + b / a))) ** 2)
        g = f * b / a

        dt_yp = (((a - 2) ** 2 + 8 * (b + c / y_p)) ** 0.5 + (a - 2)) / (2 * (b + c / y_p))
        dt_pt = y_p * (a - f) / (c + y_p * (b - g))
        dt_te = (2 + b / a) / (3 * b / a)

        if dt <= dt_yp:
            collapse_range = 'yield'
        elif dt_yp < dt < dt_pt:
            collapse_range = 'plastic'
        elif dt_pt < dt < dt_te:
            collapse_range = 'transition'
        else:
            collapse_range = 'elastic'

        pressure = p_collap(dt, y_p, collapse_range, a, b, c, f, g)
    else:
        pressure = 0

    return pressure


def p_collap(dt, y_p, c_range, a, b, c, f, g):
    if c_range == 'yield':
        p_collapse = 2 * y_p * (dt - 1) / dt**2
    elif c_range == 'plastic':
        p_collapse = y_p * ((a/dt) - b) - c
    elif c_range == 'transition':
        p_collapse = y_p * ((f/dt) - g)
    else:
        p_collapse = 46.96 * 1e6 / (dt * (dt-1)**2)

    return p_collapse



def convert_unit(value, unit_from="ft", unit_to="m"):

    result = 0
    conversion = False

    if type(value) != list:
        value = [value]

    # Length
    if (unit_from == "ft") and (unit_to == "m"):
        conversion = True
        result = [x / 3.281 for x in value]

    if (unit_from == "m") and (unit_to == "ft"):
        conversion = True
        result = [x * 3.281 for x in value]

    if (unit_from == "in") and (unit_to == "m"):
        conversion = True
        result = [x / 39.37 for x in value]

    if (unit_from == "m") and (unit_to == "in"):
        conversion = True
        result = [x * 39.37 for x in value]

    # AREA
    if (unit_from == "in2") and (unit_to == "m2"):
        conversion = True
        result = [x/1550 for x in value]

    if (unit_from == "m2") and (unit_to == "in2"):
        conversion = True
        result = [x*1550 for x in value]

    # Pressure
    if (unit_from == "Pa") and (unit_to == "bar"):
        conversion = True
        result = [x/1e5 for x in value]

    if (unit_from == "bar") and (unit_to == "Pa"):
        conversion = True
        result = [x*1e5 for x in value]

    if (unit_from == "Pa") and (unit_to == "psi"):
        conversion = True
        result = [x/6895 for x in value]

    if (unit_from == "psi") and (unit_to == "Pa"):
        conversion = True
        result = [x/6895 for x in value]

    if (unit_from == "N/mm2") and (unit_to == "bar"):
        conversion = True
        result = [x*10 for x in value]

    if (unit_from == "bar") and (unit_to == "N/mm2"):
        conversion = True
        result = [x/10 for x in value]

    if (unit_from == "psi") and (unit_to == "bar"):
        conversion = True
        result = [x/14.504 for x in value]

    if (unit_from == "bar") and (unit_to == "psi"):
        conversion = True
        result = [x*14.504 for x in value]

    # Density
    if (unit_from == "kg/m3") and (unit_to == "sg"):
        conversion = True
        result = [x/1000 for x in value]

    if (unit_from == "sg") and (unit_to == "kg/m3"):
        conversion = True
        result = [x*1000 for x in value]

    if (unit_from == "lb/in3") and (unit_to == "sg"):
        conversion = True
        result = [x*27.68 for x in value]

    if (unit_from == "sg") and (unit_to == "lb/in3"):
        conversion = True
        result = [x/27.68 for x in value]

    # Force
    if (unit_from == "kN") and (unit_to == "lbf"):
        conversion = True
        result = [x * 1000 / 4.448 for x in value]

    if (unit_from == "lbf") and (unit_to == "kN"):
        conversion = True
        result = [x * 4.448 / 1000 for x in value]

    if len(result) == 1:
        result = result[0]

    if not conversion:
        print("No conversion has been done")

    return result



import matplotlib.pyplot as plt
import plotly.graph_objects as go
from numpy import array, interp


def create_pyplot_figure(csg):

    fig, ax = plt.subplots()

    # Plotting VME
    ax.plot(csg.ellipse[0], csg.ellipse[1], 'r', label='Triaxial ' + str(csg.design_factor['vme']),
            linestyle='dashed')
    ax.plot(csg.ellipse[0], csg.ellipse[2], 'r', linestyle='dashed')

    # Plotting API limits
    ax.plot(csg.api_lines[0], csg.api_lines[1], 'k', label='API')

    # Plotting connections limits
    # for compression
    ax.plot([csg.conn_limits[0]/1000]*2,
            [csg.limits['collapseDF']/1000, csg.limits['burstDF']/1000],
            '0.55',
            label='Connection', linestyle='dashed')

    # for tension
    ax.plot([csg.conn_limits[1] / 1000] * 2,
            [csg.limits['burstDF'] / 1000,
             interp(csg.conn_limits[1], csg.collapse_curve[0], csg.collapse_curve[1]) / 1000],
            '0.55', linestyle='dashed')

    # Plotting Loads
    for load in csg.loads:
        ax.plot(array(load['axialForce']) / 1000, array(load['diffPressure']) / 1000, label=load['description'])

    ax.set_xlabel('Axial Force, klb-f')
    ax.set_ylabel('Pressure Difference, ksi')
    ax.ticklabel_format(style='plain')
    ax.legend(loc='lower right', fontsize='x-small')
    ax.grid()

    return fig


def create_plotly_figure(csg):
    fig = go.Figure()

    # Plotting VME
    triaxial_x = csg.ellipse[0] + csg.ellipse[0][::-1]
    triaxial_y = csg.ellipse[1] + csg.ellipse[2][::-1]
    fig.add_trace(go.Scatter(x=triaxial_x, y=triaxial_y, line={'color': 'red', 'dash': 'dash'},
                             name='Triaxial ' + str(csg.design_factor['vme'])))

    # Plotting API limits
    fig.add_trace(go.Scatter(x=array(csg.api_lines[0]) / 1000, y=array(csg.api_lines[1]) / 1000,
                             line={'color': 'black'}, name='API'))

    # Plotting connections limits
    fig.add_trace(go.Scatter(x=[csg.conn_limits[0] / 1000] * 2 + [None] + [csg.conn_limits[1] / 1000] * 2,
                             y=[csg.limits['collapseDF'] / 1000, csg.limits['burstDF'] / 1000] + [None] +
                               [csg.limits['burstDF'] / 1000,
                                interp(csg.conn_limits[1], csg.collapse_curve[0], csg.collapse_curve[1]) / 1000],
                             line={'color': 'gray', 'dash': 'dash'}, name='Connection', mode='lines'))

    # Plotting Loads
    for load in csg.loads:
        fig.add_trace(go.Scatter(x=array(load['axialForce']) / 1000, y=array(load['diffPressure']) / 1000,
                                 name=load['description']))

    fig.update_layout(
        xaxis_title='Axial Force, kips',
        yaxis_title='Pressure Difference, ksi')

    return fig



from ..unit_converter import convert_unit
g = 9.81        # gravity constant, [m/s2]


def inside_full(tvd, rho_fluid):

    rho_fluid = convert_unit(rho_fluid, unit_from="sg", unit_to="kg/m3")

    p_int = [g * rho_fluid * x for x in tvd]

    return p_int


def partial_evacuation(tvd, rho_mud, tvd_zone, p_zone):

    rho_mud = convert_unit(rho_mud, unit_from="sg", unit_to="kg/m3")
    tvd_mud_droplevel = tvd_zone - p_zone / (g * rho_mud)

    p_int = []
    for x in tvd:
        if x <= tvd_mud_droplevel:
            p_int.append(0)
        else:
            p_int = g * rho_mud * (x - tvd_mud_droplevel)

    return p_int


def full_evacuation(tvd):
    p_int = [0] * len(tvd)

    return p_int



from .forces import *


def running(trajectory, nominal_weight, od_csg, id_csg, shoe_depth, tvd_fluid, rho_fluid, v_avg, e,
            fric=0.24, a=1.5):
    """
    Calculate axial load during running
    :param trajectory: wellpath object
    :param nominal_weight: weight per unit length, kg/m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param shoe_depth: measured depth at shoe, m
    :param tvd_fluid: list - reference tvd of fluid change
    :param rho_fluid: list - downwards sorted fluids densities
    :param v_avg: average running speed, m/s
    :param e: pipe Young's modulus, bar
    :param fric: sliding friction factor pipe - wellbore
    :param a: ratio of maximum running speed to average running speed
    :return: total axial force profile, kN
    """

    f_w = air_weight(trajectory.tvd, nominal_weight)
    f_bu = buoyancy_force(trajectory.tvd, od_csg, id_csg, tvd_fluid, rho_fluid, tvd_fluid, rho_fluid)
    f_sh = shock_load(trajectory.tvd, v_avg, od_csg, id_csg, nominal_weight, e, a)
    f_d = drag(trajectory, od_csg, id_csg, shoe_depth, nominal_weight, tvd_fluid, rho_fluid, fric)
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [x1 - x2 + x3 - x4 + x5 for x1, x2, x3, x4, x5 in zip(f_w, f_bu, f_sh, f_d, f_be)]

    return force


def pulling(trajectory, nominal_weight, od_csg, id_csg, shoe_depth, tvd_fluid, rho_fluid, v_avg, e,
            fric=0.24, a=1.5, f_ov=0):
    """
    Calculate axial load during pulling
    :param trajectory: wellpath object
    :param nominal_weight: weight per unit length, kg/m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param shoe_depth: measured depth at shoe, m
    :param tvd_fluid: list - reference tvd of fluid change
    :param rho_fluid: list - downwards sorted fluids densities
    :param v_avg: average running speed, m/s
    :param e: pipe Young's modulus, bar
    :param fric: sliding friction factor pipe - wellbore
    :param a: ratio of maximum running speed to average running speed
    :param f_ov: overpull force (often during freeing of stuck pipe), kN.
    :return: total axial force profile, kN
    """

    f_w = air_weight(trajectory.tvd, nominal_weight)
    f_bu = buoyancy_force(trajectory.tvd, od_csg, id_csg, tvd_fluid, rho_fluid, tvd_fluid, rho_fluid)
    f_sh = shock_load(trajectory.tvd, v_avg, od_csg, id_csg, nominal_weight, e, a)
    f_d = drag(trajectory, od_csg, id_csg, shoe_depth, nominal_weight, tvd_fluid, rho_fluid, fric, 'hoisting')
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [x1 - x2 + x3 + x4 + f_ov + x5 for x1, x2, x3, x4, x5 in zip(f_w, f_bu, f_sh, f_d, f_be)]

    return force


def fluid_filled(trajectory, nominal_weight, od_csg, id_csg, rho_fluid_ext, rho_fluid_int, e):
    """
    Calculate axial load when casing is filled with a specific fluid.
    :param trajectory: wellpath object
    :param nominal_weight: weight per unit length, kg/m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param rho_fluid_ext: list - downwards sorted fluids densities outside, sg
    :param rho_fluid_int: list - downwards sorted fluids densities inside, sg
    :param e: pipe Young's modulus, bar
    :return: total axial force profile, kN
    """

    f_w = air_weight(trajectory.tvd, nominal_weight)
    f_bu = buoyancy_force(trajectory.tvd, od_csg, id_csg, [], [rho_fluid_ext], [], [rho_fluid_int])
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [x1 - x2 + x3 for x1, x2, x3 in zip(f_w, f_bu, f_be)]

    return force


def cementation(trajectory, nominal_weight, od_csg, id_csg, rho_cement, rho_fluid, e, f_pre=0):
    """
    Calculate axial load during cementing
    :param trajectory: wellpath object
    :param nominal_weight: weight per unit length, kg/m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param rho_cement: cement density, sg
    :param rho_fluid: displacement fluid density, sg
    :param e: pipe Young's modulus, bar
    :param f_pre: pre-loading force applied to the casing string if necessary, kN
    :return: total axial force profile, kN
    """

    f_w = air_weight(trajectory.tvd, nominal_weight)
    f_bu = buoyancy_force(trajectory.tvd, od_csg, id_csg, [], [rho_cement], [], [rho_fluid])
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [x1 - x2 + f_pre + x3 for x1, x2, x3 in zip(f_w, f_bu, f_be)]

    return force


def green_cement(trajectory, nominal_weight, od_csg, id_csg, rho_cement, rho_fluid_int, e, f_pre=0, f_h=0):
    """
    Calculate axial load during green cement pressure test
    :param trajectory: wellpath object
    :param nominal_weight: weight per unit length, kg/m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param rho_cement: cement density, sg
    :param rho_fluid_int: inside fluid density, sg
    :param e: pipe Young's modulus, bar
    :param f_pre: pre-loading force applied to the casing string if necessary, kN
    :param f_h: pressure testing force, kN
    :return: total axial force profile, kN
    """

    f_w = air_weight(trajectory.tvd, nominal_weight)
    f_bu = buoyancy_force(trajectory.tvd, od_csg, id_csg, [], [rho_cement], [], [rho_fluid_int])
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [x1 - x2 + f_h + f_pre + x3 for x1, x2, x3 in zip(f_w, f_bu, f_be)]

    return force


def production(trajectory, md_toc, od_csg, id_csg, rho_fluid_int, rho_fluid_ext, e, poisson=0.3, f_setting=0):
    """
    Calculate axial load during production
    :param trajectory: wellpath object
    :param md_toc: md at top of cement, m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param rho_fluid_int: fluid density in the casing string after installation
    :param rho_fluid_ext: density of annular fluid after installation
    :param e: pipe Young's modulus, bar
    :param poisson: Poisson’s ratio
    :param f_setting: hang off force of the casing string on slips or hangers, kN
    :return: total axial force profile, kN
    """

    f_bl = ballooning(trajectory.md, md_toc, od_csg, id_csg, rho_fluid_int, rho_fluid_ext, poisson)
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [f_setting + x1 + x2 for x1, x2 in zip(f_bl, f_be)]

    return force


def injection(trajectory, md_toc, od_csg, id_csg, rho_fluid_int, rho_fluid_ext, e, t_k, temp, alpha=17e-6, poisson=0.3,
              f_setting=0):
    """
    Calculate axial load during injection
    :param trajectory: wellpath object
    :param md_toc: md at top of cement, m
    :param od_csg: pipe outer diameter, in
    :param id_csg: pipe inner diameter, in
    :param rho_fluid_int: fluid density in the casing string after installation
    :param rho_fluid_ext: density of annular fluid after installation
    :param e: pipe Young's modulus, bar
    :param t_k: max. wellhead temperature during production, °C
    :param temp: dict with temp values (°C) at seabed and a target
    :param alpha: thermal expansion coefficient, 1/°C
    :param poisson: Poisson’s ratio
    :param f_setting: hang off force of the casing string on slips or hangers, kN
    :return: total axial force profile, kN
    """

    from .forces import thermal_load

    f_th = thermal_load(trajectory, od_csg, id_csg, t_k, temp, alpha, e)
    f_bl = ballooning(trajectory.md, md_toc, od_csg, id_csg, rho_fluid_int, rho_fluid_ext, poisson)
    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)

    force = [f_setting + x1 + x2 + x3 for x1, x2, x3 in zip(f_bl, f_th, f_be)]

    return force


def pressure_test(trajectory, whp, effective_diameter, od_csg,  e):
    """
    Calculate axial load during green cement pressure test
    :param trajectory: wellpath object
    :param whp: wellhead pressure, bar
    :param od_csg: pipe outer diameter, in
    :param effective_diameter: pipe inner diameter, in
    :param e: pipe Young's modulus, bar
    :return: total axial force profile, kN
    """

    f_be = bending(od_csg, trajectory.dls, trajectory.info['dlsResolution'], e)
    f_h = pressure_testing(trajectory.tvd, whp, effective_diameter)

    force = [x1 + x2 for x1, x2 in zip(f_h, f_be)]

    return force



def drilling_losses(tvd, rho_mud, tvd_zone, p_zone):
    from .pressure_internal import partial_evacuation
    from .pressure_external import onefluid_behindcasing

    p_int = partial_evacuation(tvd, rho_mud, tvd_zone, p_zone)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def plug_cementation_onefluid_behindcasing(tvd, rho_fluid, rho_mud):
    from .pressure_internal import inside_full
    from .pressure_external import onefluid_behindcasing

    p_int = inside_full(tvd, rho_fluid)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def plug_cementation_morefluids_behindcasing(tvd, rho_fluid, tvd_fluid):
    from .pressure_internal import inside_full
    from .pressure_external import morefluids_behindcasing

    p_int = inside_full(tvd, rho_fluid)
    p_ext = morefluids_behindcasing(tvd, rho_fluid, tvd_fluid)

    pressure_differential = p_int - p_ext

    return pressure_differential


def drill_stem_test_partialevacuation(tvd, rho_mud, tvd_zone, p_zone):
    from .pressure_internal import partial_evacuation
    from .pressure_external import onefluid_behindcasing

    p_int = partial_evacuation(tvd, rho_mud, tvd_zone, p_zone)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def drill_stem_test_fullevacuation(tvd, rho_mud):
    from .pressure_internal import full_evacuation
    from .pressure_external import onefluid_behindcasing

    p_int = full_evacuation(tvd)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = p_int - p_ext

    return pressure_differential


def production_fullevacuation(tvd, rho_mud):
    from .pressure_internal import full_evacuation
    from .pressure_external import onefluid_behindcasing

    p_int = full_evacuation(tvd)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential


def injection_partialevacuation(tvd, tvd_perf, rho_inj, p_inj, p_zone, tvd_influencedzone, rho_fluid_before_cementation,
                                p_fric, rho_form, rho_fluid_above_packer):
    from .pressure_internal import partial_evacuation
    from .pressure_external import injection

    p_int = partial_evacuation(tvd, rho_fluid_above_packer, tvd_perf, p_zone)
    p_ext = injection(tvd, tvd_perf, p_inj, rho_inj, tvd_influencedzone, rho_fluid_before_cementation, p_fric, rho_form)

    pressure_differential = p_int - p_ext

    return pressure_differential


def injection_fulllevacuation(tvd, tvd_perf, rho_inj, p_inj, tvd_influencedzone, rho_fluid, p_fric, rho_form):
    from .pressure_internal import full_evacuation
    from .pressure_external import injection

    p_int = full_evacuation(tvd)
    p_ext = injection(tvd, tvd_perf, p_inj, rho_inj, tvd_influencedzone, rho_fluid, p_fric, rho_form)

    pressure_differential = p_int - p_ext

    return pressure_differential


def fluid_filled(tvd, rho_mud_new, rho_mud):
    """
    Calculate differential pressure profile when the entire casing string is filled with gas.
    :param tvd: list - true vertical depth, m
    :param rho_mud_new: new mud density, sg
    :param rho_mud: mud density, sg
    :return: differential pressure profile, Pa
    """

    from .pressure_internal import inside_full
    from .pressure_external import onefluid_behindcasing

    p_int = inside_full(tvd, rho_mud_new)
    p_ext = onefluid_behindcasing(tvd, rho_mud)

    pressure_differential = [x - y for x, y in zip(p_int, p_ext)]

    return pressure_differential
